/* #includes */ /*{{{C}}}*//*{{{*/
#include <assert.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "cpmdir.h"
#include "cpmfs.h"
/*}}}*/
/* #defines */ /*{{{*/
#undef CPMFS_DEBUG
/*}}}*/

extern char **environ;

#define INTBITS (sizeof(int)*8)

const char *boo;

/* file name conversions */ 
/* splitFilename      -- split file name into name and extension */ /*{{{*/
static int splitFilename(const char *fullname, int type, char *name, char *ext, int *user) 
{
  int i,j;

  assert(fullname!=(const char*)0);
  assert(name!=(char*)0);
  assert(ext!=(char*)0);
  assert(user!=(int*)0);
  memset(name,' ',8);
  memset(ext,' ',3);
  *user=0;
  if (isalpha(fullname[0])) /* try to match drive and user number */ /*{{{*/
  {
    if (isdigit(fullname[1]))
    {
      if (isdigit(fullname[2]) && fullname[3]==':')
      {
        *user=(fullname[1]-'0')*10+(fullname[2]-'0');
        fullname+=4;
      }
      else if (fullname[2]==':')
      {
        *user=(fullname[1]-'0');
        fullname+=3;
      }
    }
    else if (fullname[1]==':') fullname+=2;
  }
  /*}}}*/
  if ((fullname[0]=='\0') || (type==CPMFS_DR22 && *user>=16) || (type==CPMFS_P2DOS && *user>=32))
  {
    boo="illegal CP/M filename";
    return -1;
  }
  for (i=0; i<8 && fullname[i] && fullname[i]!='.'; ++i) if (!ISFILECHAR(i,fullname[i]))
  {
    boo="illegal CP/M filename";
    return -1;
  }
  else name[i]=toupper(fullname[i]);
  if (fullname[i]=='.')
  {
    ++i;
    for (j=0; j<3 && fullname[i]; ++i,++j) if (!ISFILECHAR(1,fullname[i]))
    {
      boo="illegal CP/M filename";
      return -1;
    }
    else ext[j]=toupper(fullname[i]);
    if (i==1 && j==0)
    {
      boo="illegal CP/M filename";
      return -1;
    }
  }
  return 0;
}
/*}}}*/
/* isMatching         -- do two file names match?                */ /*{{{*/
static int isMatching(int user1, const char *name1, const char *ext1, int user2, const char *name2, const char *ext2)
{
  int i;

  assert(name1!=(const char*)0);
  assert(ext1!=(const char*)0);
  assert(name2!=(const char*)0);
  assert(ext2!=(const char*)0);
  if (user1!=user2) return 0;
  for (i=0; i<8; ++i) if ((name1[i]&0x7f)!=(name2[i]&0x7f)) return 0;
  for (i=0; i<3; ++i) if ((ext1[i]&0x7f)!=(ext2[i]&0x7f)) return 0;
  return 1;
}
/*}}}*/

/* allocation vector bitmap functions */
/* alvInit            -- init allocation vector                  */ /*{{{*/
static void alvInit(const struct cpmSuperBlock *d)
{
  int i,j,offset,block;

  assert(d!=(const struct cpmSuperBlock*)0);
  /* clean bitmap */ /*{{{*/
  memset(d->alv,0,d->alvSize);
  /*}}}*/
  /* mark directory blocks as used */ /*{{{*/
  *d->alv=(1<<((d->maxdir*32+d->blksiz-1)/d->blksiz))-1;
  /*}}}*/
  for (i=0; i<d->maxdir; ++i) /* mark file blocks as used */ /*{{{*/
  {
    if (d->dir[i].status!=(char)0xe5 && d->dir[i].status<=32)
    {
      for (j=0; j<16; ++j)
      {
        block = d->dir[i].pointers[j]&0xff;
        if (d->size>=256) block+=((d->dir[i].pointers[++j]&0xff)<<8);
        if (block)
        {
          offset=block/INTBITS;
          d->alv[offset]|=(1<<block%INTBITS);
        }
      }
    }
  }
  /*}}}*/
}
/*}}}*/
/* allocBlock         -- allocate a new disk block               */ /*{{{*/
static int allocBlock(const struct cpmSuperBlock *drive)
{
  int i,j,bits,block;

  assert(drive!=(const struct cpmSuperBlock*)0);
  for (i=0; i<drive->alvSize; ++i)
  {
    for (j=0,bits=drive->alv[i]; j<INTBITS; ++j)
    {
      if ((bits&1)==0)
      {
        block=i*INTBITS+j;
        if (block>=drive->size)
        {
          boo="device full";
          return -1;
        }
        drive->alv[i] |= (1<<j);
        return block;
      }
      bits >>= 1;
    }
  }
  boo="device full";
  return -1;
}
/*}}}*/

/* physical sector I/O */
/* readSector         -- read a physical sector                  */ /*{{{*/
static int readSector(const struct cpmSuperBlock *drive, int track, int sector, char *buf)
{
  int res;

  assert(sector>=0);
  assert(sector<drive->sectrk);
  assert(track>=0);
  assert(track<drive->tracks);
  if (lseek(drive->fd,(off_t)(sector+track*drive->sectrk)*drive->secLength,SEEK_SET)==-1) 
  {
    boo=strerror(errno);
    return -1;
  }
  if ((res=read(drive->fd, buf, drive->secLength)) != drive->secLength) 
  {
    if (res==-1)
    {
      boo=strerror(errno);
      return -1;
    }
    else memset(buf+res,0,drive->secLength-res); /* hit end of disk image */
  }
  return 0;
}
/*}}}*/
/* writeSector        -- write physical sector                   */ /*{{{*/
static int writeSector(const struct cpmSuperBlock *drive, int track, int sector, const char *buf)
{
  assert(sector>=0);
  assert(sector<drive->sectrk);
  assert(track>=0);
  assert(track<drive->tracks);
  if (lseek(drive->fd,(off_t)(sector+track*drive->sectrk)*drive->secLength, SEEK_SET)==-1) 
  {
    boo=strerror(errno);
    return -1;
  }
  if (write(drive->fd, buf, drive->secLength) == drive->secLength) return 0;
  boo=strerror(errno);
  return -1;
}
/*}}}*/

/* logical block I/O */
/* readBlock          -- read a (partial) block                  */ /*{{{*/
static int readBlock(const struct cpmSuperBlock *d, int blockno, char *buffer, int start, int end)
{
  int sect, track, counter;

  assert(blockno>=0);
  assert(blockno<d->size);
  assert(buffer!=(char*)0);
  if (end<0) end=d->blksiz/d->secLength-1;
  sect=(blockno*(d->blksiz/d->secLength)+ d->sectrk*d->boottrk)%d->sectrk;
  track=(blockno*(d->blksiz/d->secLength)+ d->sectrk*d->boottrk)/d->sectrk;
  for (counter=0; counter<=end; ++counter)
  {
    if (counter>=start && readSector(d,track,d->skewtab[sect],buffer+(d->secLength*counter))==-1) return -1;
    ++sect;
    if (sect>=d->sectrk) 
    {
      sect = 0;
      ++track;
    }
  }
  return 0;
}
/*}}}*/
/* writeBlock         -- write a (partial) block                 */ /*{{{*/
static int writeBlock(const struct cpmSuperBlock *d, int blockno, const char *buffer, int start, int end)
{
  int sect, track, counter;

  assert(blockno>=0);
  assert(blockno<d->size);
  assert(buffer!=(const char*)0);
  if (end < 0) end=d->blksiz/d->secLength-1;
  sect = (blockno*(d->blksiz/d->secLength))%d->sectrk;
  track = (blockno*(d->blksiz/d->secLength))/d->sectrk+d->boottrk;
  for (counter = 0; counter<=end; ++counter)
  {
    if (counter>=start && writeSector(d,track,d->skewtab[sect],buffer+(d->secLength*counter))==-1) return -1;
    ++sect;
    if (sect>=d->sectrk) 
    {
      sect=0;
      ++track;
    }
  }
  return 0;
}
/*}}}*/

/* directory management */
/* readPhysDirectory  -- read directory from drive               */ /*{{{*/
static int readPhysDirectory(const struct cpmSuperBlock *drive)
{
  int i,blocks,entry;

  blocks=(drive->maxdir*32+drive->blksiz-1)/drive->blksiz;
  entry=0;
  for (i=0; i<blocks; ++i) 
  {
    if (readBlock(drive,i,(char*)(drive->dir+entry),0,-1)==-1) return -1;
    entry+=(drive->blksiz/32);
  }
  return 0;
}
/*}}}*/
/* writePhysDirectory -- write directory to drive                */ /*{{{*/
static int writePhysDirectory(const struct cpmSuperBlock *drive)
{
  int i,blocks,entry;

  blocks=(drive->maxdir*32+drive->blksiz-1)/drive->blksiz;
  entry=0;
  for (i=0; i<blocks; ++i) 
  {
    if (writeBlock(drive,i,(char*)(drive->dir+entry),0,-1)==-1) return -1;
    entry+=(drive->blksiz/32);
  }
  return 0;
}
/*}}}*/
/* findFileExtent     -- find first/next extent for a file       */ /*{{{*/
static int findFileExtent(const struct cpmSuperBlock *sb, int user, const char *name, const char *ext, int start, int extno)
{
  boo="file already exists";
  for (; start<sb->maxdir; ++start)
  {
    if
    (
      (sb->dir[start].status&0xff)<=31 
      && (extno==-1 || (EXTENT(sb->dir[start].extnol,sb->dir[start].extnoh)/sb->extents)==(extno/sb->extents))
      && isMatching(user,name,ext,sb->dir[start].status,sb->dir[start].name,sb->dir[start].ext)
    ) return start;
  }
  boo="file not found";
  return -1;
}
/*}}}*/
/* findFreeExtent     -- find first free extent                  */ /*{{{*/
static int findFreeExtent(const struct cpmSuperBlock *drive)
{
  int i;

  for (i=0; i<drive->maxdir; ++i) if (drive->dir[i].status==(char)0xe5) return (i);
  boo="directory full";
  return -1;
}
/*}}}*/
/* updateTimeStamps   -- convert time stamps to CP/M format      */ /*{{{*/
static void updateTimeStamps(const struct cpmInode *ino, int extent)
{
  struct PhysDirectoryEntry *date;

  if (!S_ISREG(ino->mode)) return;
  if (ino->sb->type==CPMFS_P2DOS && (date=ino->sb->dir+(extent|3))->status==0x21)
  {
    struct tm *t;
    int i;
    int c_min,c_hour,c_days,u_min,u_hour,u_days;

#ifdef CPMFS_DEBUG
    fprintf(stderr,"CPMFS: updating time stamps for inode %d (%d)\n",extent,extent&3);
#endif
    /* compute ctime */ /*{{{*/
    t=localtime(&ino->ctime);
    c_min=((t->tm_min/10)<<4)|(t->tm_min%10);
    c_hour=((t->tm_hour/10)<<4)|(t->tm_hour%10);
    for (i=1978,c_days=0; i < 1900 + t->tm_year; ++i)
    {
      c_days+=365;
      if (i%4==0 && (i%100!=0 || i%400==0)) ++c_days;
    }
    c_days += t->tm_yday + 1;
    /*}}}*/
    /* compute mtime */ /*{{{*/
    t=localtime(&ino->mtime);
    u_min=((t->tm_min/10)<<4)|(t->tm_min%10);
    u_hour=((t->tm_hour/10)<<4)|(t->tm_hour%10);
    for (i=1978,u_days=0; i < 1900 + t->tm_year; ++i)
    {
      u_days+=365;
      if (i%4==0 && (i%100!=0 || i%400==0)) ++u_days;
    }
    u_days += t->tm_yday + 1;
    /*}}}*/
    switch (extent&3)
    {
      case 0: /* first entry */ /*{{{*/
      {
        date->name[0]=u_days&0xff; date->name[1]=u_days>>8;
        date->name[2]=u_hour;
        date->name[3]=u_min;
        date->name[4]=c_days&0xff; date->name[5]=c_days>>8;
        date->name[6]=c_hour;
        date->name[7]=c_min;
        break;
      }
      /*}}}*/
      case 1: /* second entry */ /*{{{*/
      {
        date->ext[2]=u_days&0xff; date->extnol=u_days>>8;
        date->lrc=u_hour;
        date->extnoh=u_min;
        date->blkcnt=c_days&0xff; date->pointers[0]=c_days>>8;
        date->pointers[1]=c_hour;
        date->pointers[2]=c_min;
        break;
      }
      /*}}}*/
      case 2: /* third entry */ /*{{{*/
      {
        date->pointers[5]=u_days&0xff; date->pointers[6]=u_days>>8;
        date->pointers[7]=u_hour;
        date->pointers[8]=u_min;
        date->pointers[9]=c_days&0xff; date->pointers[10]=c_days>>8;
        date->pointers[11]=c_hour;
        date->pointers[12]=c_min;
        break;
      }
      /*}}}*/
    }
  }
}
/*}}}*/

/* getformat          -- get DPB and init in-core data for drive */ /*{{{*/
int getformat(const char *format, int type, struct cpmSuperBlock *d, struct cpmInode *root)
{
  /* read format specific values from diskdefs file */ /*{{{*/
  {
    char buf[128],str[32];
    FILE *fp;
		  
    if ((fp=fopen(DISKDEFS,"r"))==(FILE*)0 && (fp=fopen("diskdefs","r"))==(FILE*)0)
    {
      fprintf(stderr,"%s: Neither " DISKDEFS " nor diskdefs could be opened.\n",cmd);
      exit(1);
    }
    while (fgets(buf,sizeof(buf),fp)!=(char*)0)
    {
      if (buf[0]=='#' || buf[0]=='\n') continue;
      if (sscanf(buf,"%s %d %d %d %d %d %d %d\n",str,&(d->secLength),&(d->tracks),&(d->sectrk),&(d->blksiz),&(d->maxdir),&(d->skew),&(d->boottrk))!=8)
      {
        fprintf(stderr,"%s: invalid line for format %s\n",cmd,format);
        continue;
      }
      if (strcmp(str,format)==0) break;
    }
    fclose(fp);
    if (strcmp(str,format))
    {
      fprintf(stderr,"%s: unknown format %s\n",cmd,format);
      exit(1);
    }
    d->size=(d->secLength*d->sectrk*(d->tracks-d->boottrk))/d->blksiz;
    d->extents=((d->size>=256 ? 8 : 16)*d->blksiz)/16384;
    d->type=type;
  }
  /*}}}*/
  /* generate skew table */ /*{{{*/
  {
    int	i,j,k;

    if (( d->skewtab = malloc(d->sectrk*sizeof(int))) == (int*)0) 
    {
      fprintf(stderr,"%s: can not allocate memory for skew sector table\n",cmd);
      exit(1);
    }
    for (i=j=0; i<d->sectrk; ++i,j=(j+d->skew)%d->sectrk)
    {
      while (1)
      {
        for (k=0; k<i && d->skewtab[k]!=j; ++k);
        if (k<i) j=(j+1)%d->sectrk;
        else break;
      }
      d->skewtab[i]=j;
    }
  }
  /*}}}*/
  /* initialise allocation vector bitmap */ /*{{{*/
  {
    d->alvSize=((d->secLength*d->sectrk*(d->tracks-d->boottrk))/d->blksiz+INTBITS-1)/INTBITS;
    if ((d->alv=malloc(d->alvSize*sizeof(int)))==(int*)0) 
    {
      boo="out of memory";
      return -1;
    }
  }
  /*}}}*/
  /* allocate directory buffer */ /*{{{*/
  if ((d->dir=malloc(d->maxdir*32))==(struct PhysDirectoryEntry*)0)
  {
    boo="out of memory";
    return -1;
  }
  /*}}}*/
  if (d->fd==-1) memset(d->dir,0xe5,d->maxdir*32);
  else if (readPhysDirectory(d)==-1) return -1;
  alvInit(d);
  d->root=root;
  root->ino=d->maxdir;
  root->sb=d;
  root->mode=(S_IFDIR|0777);
  root->size=0;
  root->atime=root->mtime=root->ctime=0;
#ifdef CPMFS_DEBUG
  fprintf(stderr,"CPMFS: type ");
  switch (d->type)
  {
    case CPMFS_DR22: fprintf(stderr,"DR 2.2\n"); break;
    case CPMFS_P2DOS: fprintf(stderr,"P2DOS\n"); break;
  }
#endif
  return 0;
}
/*}}}*/

/* cpmNamei           -- map name to inode                       */ /*{{{*/
int cpmNamei(const struct cpmInode *dir, const char *filename, struct cpmInode *i)
{
  /* variables */ /*{{{*/
  int user;
  char name[8],extension[3];
  struct PhysDirectoryEntry *date;
  char **old_environ;
  static char gmt0[]="TZ=GMT0";
  static char *gmt_env[]={ gmt0, (char*)0 };
  int highestExtno,highestExt=-1,lowestExtno,lowestExt=-1;
  /*}}}*/

  if (!S_ISDIR(dir->mode))
  {
    boo="No such file";
    return -1;
  }
  if (strcmp(filename,".")==0 || strcmp(filename,"..")==0)
  {
    *i=*dir;
    return 0;
  }
  if (splitFilename(filename,dir->sb->type,name,extension,&user)==-1) return -1;
  /* find highest and lowest extent */ /*{{{*/
  {
    int extent;

    i->size=0;
    extent=-1;
    highestExtno=-1;
    lowestExtno=2049;
    while ((extent=findFileExtent(dir->sb,user,name,extension,extent+1,-1))!=-1)
    {
      int extno=EXTENT(dir->sb->dir[extent].extnol,dir->sb->dir[extent].extnoh);

      if (extno>highestExtno)
      {
        highestExtno=extno;
        highestExt=extent;
      }
      if (extno<lowestExtno)
      {
        lowestExtno=extno;
        lowestExt=extent;
      }
    }
  }
  /*}}}*/
  if (highestExtno==-1) return -1;
  /* calculate size */ /*{{{*/
  {
    int block;

    i->size=highestExtno*16384;
    if (dir->sb->size<256) for (block=15; block>=0; --block)
    {
      if (dir->sb->dir[highestExt].pointers[block]) break;
    }
    else for (block=7; block>=0; --block)
    {
      if (dir->sb->dir[highestExt].pointers[2*block] || dir->sb->dir[highestExt].pointers[2*block+1]) break;
    }
    if (dir->sb->dir[highestExt].blkcnt) i->size+=((dir->sb->dir[highestExt].blkcnt&0xff)-1)*128;
    i->size+=dir->sb->dir[highestExt].lrc ? (dir->sb->dir[highestExt].lrc&0xff) : 128;
#ifdef CPMFS_DEBUG
    fprintf(stderr,"cpmNamei: size=%ld\n",(long)i->size);
#endif
  }
  /*}}}*/
  i->ino=lowestExt;
  i->mode=S_IFREG;
  i->sb=dir->sb;
  old_environ=environ;
  environ=gmt_env; /* for mktime() */
  if ((date=dir->sb->dir+(lowestExt|3))->status==0x21) /* set time stamps */ /*{{{*/
  {
    /* variables */ /*{{{*/
    int u_days=0,u_hour=0,u_min=0;
    int c_days=0,c_hour=0,c_min=0;
    struct tm tms;
    /*}}}*/

    switch (lowestExt&3)
    {
      case 0: /* first entry of the four */ /*{{{*/
      {
        u_days=date->name[0]+(date->name[1]<<8);
        u_hour=date->name[2];
        u_min=date->name[3];
        c_days=date->name[4]+(date->name[5]<<8);
        c_hour=date->name[6];
        c_min=date->name[7];
        break;
      }
      /*}}}*/
      case 1: /* second entry */ /*{{{*/
      {
        u_days=date->ext[2]+(date->extnol<<8);
        u_hour=date->lrc;
        u_min=date->extnoh;
        c_days=date->blkcnt+(date->pointers[0]<<8);
        c_hour=date->pointers[1];
        c_min=date->pointers[2];
        break;
      }
      /*}}}*/
      case 2: /* third one */ /*{{{*/
      {
        u_days=date->pointers[5]+(date->pointers[6]<<8);
        u_hour=date->pointers[7];
        u_min=date->pointers[8];
        c_days=date->pointers[9]+(date->pointers[10]<<8);
        c_hour=date->pointers[11];
        c_min=date->pointers[12];
        break;
      }
      /*}}}*/
    }
    /* compute CP/M to UNIX time format */ /*{{{*/
    tms.tm_sec=0;
    tms.tm_min=((u_min>>4)&0xf)*10+(u_min&0xf);
    tms.tm_hour=((u_hour>>4)&0xf)*10+(u_hour&0xf);
    tms.tm_mday=1;
    tms.tm_mon=0;
    tms.tm_year=78;
    tms.tm_isdst=-1;
    i->mtime=mktime(&tms)+(u_days-1)*24*3600;
    tms.tm_min=((c_min>>4)&0xf)*10+(c_min&0xf);
    tms.tm_hour=((c_hour>>4)&0xf)*10+(c_hour&0xf);
    i->ctime=mktime(&tms)+(c_days-1)*24*3600;
    i->atime=0;
    /*}}}*/
  }
  /*}}}*/
  else i->atime=i->mtime=i->ctime=0;
  environ=old_environ;
  i->mode|=0444;
  if (!(dir->sb->dir[lowestExt].ext[0]&0x80)) i->mode|=0222;
  if (extension[0]=='C' && extension[1]=='O' && extension[2]=='M') i->mode|=0111;
  return 0;
}
/*}}}*/
/* cpmStatFS          -- statfs                                  */ /*{{{*/
void cpmStatFS(const struct cpmInode *ino, struct cpmStatFS *buf)
{
  int i;
  struct cpmSuperBlock *d;

  d=ino->sb;
  buf->f_bsize=d->blksiz;
  buf->f_blocks=(d->tracks*d->sectrk*d->secLength)/d->blksiz;
  buf->f_bfree=0;
  buf->f_bused=-(d->maxdir*32+d->blksiz-1)/d->blksiz;
  for (i=0; i<d->alvSize; ++i)
  {
      int temp,j;

      temp = *(d->alv+i);
      for (j=0; j < INTBITS; j++) 
      {
        if ((i*INTBITS+j)<d->size) if (1&temp) ++buf->f_bused; else ++buf->f_bfree;
        temp >>= 1;
      }
  }
  buf->f_bavail=buf->f_bfree;
  buf->f_files=d->maxdir;
  buf->f_ffree=0;
  for (i=0; i<d->maxdir; ++i)
  {
    if (d->dir[i].status==(char)0xe5) ++buf->f_ffree;
    if (d->dir[i].status==(char)0x21) --buf->f_files;
  }
  buf->f_namelen=11;
}
/*}}}*/
/* cpmUnlink          -- unlink                                  */ /*{{{*/
int cpmUnlink(const struct cpmInode *dir, const char *fname)
{
  int user;
  char name[8],extension[3];
  int extent;
  struct cpmSuperBlock *drive;

  if (!S_ISDIR(dir->mode))
  {
    boo="No such file";
    return -1;
  }
  drive=dir->sb;
  if (splitFilename(fname,dir->sb->type,name,extension,&user)==-1) return -1;
  if ((extent=findFileExtent(drive,user,name,extension,0,-1))==-1) return -1;
  drive->dir[extent].status=(char)0xe5;
  do
  {
    drive->dir[extent].status=(char)0xe5;
  } while ((extent=findFileExtent(drive,user,name,extension,extent+1,-1))>=0);
  if (writePhysDirectory(drive)==-1) return -1;
  alvInit(drive);
  return 0;
}
/*}}}*/
/* cpmRename          -- rename                                  */ /*{{{*/
int cpmRename(const struct cpmInode *dir, const char *old, const char *new)
{
  struct cpmSuperBlock *drive;
  int extent;
  int olduser;
  char oldname[8], oldext[3];
  int newuser;
  char newname[8], newext[3];

  if (!S_ISDIR(dir->mode))
  {
    boo="No such file";
    return -1;
  }
  drive=dir->sb;
  if (splitFilename(old,dir->sb->type, oldname, oldext,&olduser)==-1) return -1;
  if (splitFilename(new,dir->sb->type, newname, newext,&newuser)==-1) return -1;
  if ((extent=findFileExtent(drive,olduser,oldname,oldext,0,-1))==-1) return -1;
  if (findFileExtent(drive,newuser,newname, newext,0,-1)!=-1) 
  {
    boo="file already exists";
    return -1;
  }
  do 
  {
    drive->dir[extent].status=newuser;
    memcpy(drive->dir[extent].name, newname, 8);
    memcpy(drive->dir[extent].ext, newext, 3);
  } while ((extent=findFileExtent(drive,olduser,oldname,oldext,extent+1,-1))!=-1);
  if (writePhysDirectory(drive)==-1) return -1;
  return 0;
}
/*}}}*/
/* cpmOpendir         -- opendir                                 */ /*{{{*/
int cpmOpendir(struct cpmInode *dir, struct cpmFile *dirp)
{
  if (!S_ISDIR(dir->mode))
  {
    boo="No such file";
    return -1;
  }
  dirp->ino=dir;
  dirp->pos=0;
  dirp->mode=O_RDONLY;
  return 0;
}
/*}}}*/
/* cpmReaddir         -- readdir                                 */ /*{{{*/
int cpmReaddir(struct cpmFile *dir, struct cpmDirent *ent)
{
  /* variables */ /*{{{*/
  struct PhysDirectoryEntry *cur=(struct PhysDirectoryEntry*)0;
  char buf[13];
  int i;
  char *bufp;
  int hasext;
  /*}}}*/

  if (!(S_ISDIR(dir->ino->mode))) /* error: not a directory */ /*{{{*/
  {
    boo="not a directory";
    return -1;
  }
  /*}}}*/
  if (dir->pos==0) /* first entry is . */ /*{{{*/
  {
    ent->ino=dir->ino->sb->maxdir;
    ent->reclen=1;
    strcpy(ent->name,".");
  }
  /*}}}*/
  else if (dir->pos==1) /* next entry is .. */ /*{{{*/
  {
    ent->ino=dir->ino->sb->maxdir;
    ent->reclen=2;
    strcpy(ent->name,"..");
  }
  /*}}}*/
  else
  {
    /* find next entry */ /*{{{*/
    while ((dir->pos-2)<dir->ino->sb->maxdir)
    {
      int first=dir->pos-2;

      if ((cur=dir->ino->sb->dir+(dir->pos-2))->status>=0 && cur->status<=31)
      {
        /* determine first extent for the current file */ /*{{{*/
        for (i=0; i<dir->ino->sb->maxdir; ++i) if (i!=(dir->pos-2))
        {
          if (isMatching(cur->status,cur->name,cur->ext,dir->ino->sb->dir[i].status,dir->ino->sb->dir[i].name,dir->ino->sb->dir[i].ext) && EXTENT(cur->extnol,cur->extnoh)>EXTENT(dir->ino->sb->dir[i].extnol,dir->ino->sb->dir[i].extnoh)) first=i;
        }
        /*}}}*/
        if (first!=(dir->pos-2)) ++dir->pos;
        else break;
      }
      else ++dir->pos;
    }
    /*}}}*/
    if ((dir->pos-2)==dir->ino->sb->maxdir) return 0;
    ent->ino=dir->pos-2;
    /* convert file name to UNIX style */ /*{{{*/
    for (bufp=buf,i=0; i<8 && (cur->name[i]&0x7f)!=' '; ++i) *bufp++=tolower(cur->name[i]&0x7f);
    for (hasext=0,i=0; i<3 && (cur->ext[i]&0x7f)!=' '; ++i)
    {
      if (!hasext) { *bufp++='.'; hasext=1; }
      *bufp++=tolower(cur->ext[i]&0x7f);
    }
    *bufp='\0';
    /*}}}*/
    ent->reclen=strlen(buf);
    strcpy(ent->name,buf);
  }
  ent->off=dir->pos;
  ++dir->pos;
  return 1;
}
/*}}}*/
/* cpmStat            -- stat                                    */ /*{{{*/
void cpmStat(const struct cpmInode *ino, struct cpmStat *buf)
{
  buf->ino=ino->ino;
  buf->mode=ino->mode;
  buf->size=ino->size;
  buf->atime=ino->atime;
  buf->mtime=ino->mtime;
  buf->ctime=ino->ctime;
}
/*}}}*/
/* cpmOpen            -- open                                    */ /*{{{*/
int cpmOpen(struct cpmInode *ino, struct cpmFile *file, mode_t mode)
{
  if (S_ISREG(ino->mode))
  {
    if ((mode&O_WRONLY) && (ino->mode&0222)==0)
    {
      boo="permission denied";
      return -1;
    }
    file->pos=0;
    file->ino=ino;
    file->mode=mode;
    return 0;
  }
  else
  {
    boo="not a regular file";
    return -1;
  }
}
/*}}}*/
/* cpmRead            -- read                                    */ /*{{{*/
int cpmRead(struct cpmFile *file, char *buf, int count)
{
  int findext=1,findblock=1,extent=-1,block=-1,extentno=-1,got=0,nextblockpos=-1,nextextpos=-1;
  int blocksize=file->ino->sb->blksiz;
  int extcap=(file->ino->sb->size<256 ? 16 : 8)*blocksize;

  while (count>0 && file->pos<file->ino->size)
  {
    char buffer[16384];

    if (findext)
    {
      extentno=file->pos/16384;
      extent=findFileExtent(file->ino->sb,file->ino->sb->dir[file->ino->ino].status,file->ino->sb->dir[file->ino->ino].name,file->ino->sb->dir[file->ino->ino].ext,0,extentno);
      nextextpos=(file->pos/extcap)*extcap+extcap;
      findext=0;
      findblock=1;
    }
    if (findblock)
    {
      if (extent!=-1)
      {
        int start,end,ptr;

        ptr=(file->pos%extcap)/blocksize;
        if (file->ino->sb->size>=256) ptr*=2;
        block=file->ino->sb->dir[extent].pointers[ptr]&0xff;
        if (file->ino->sb->size>=256) block+=(file->ino->sb->dir[extent].pointers[ptr+1]&0xff)<<8;
        start=(file->pos%blocksize)/file->ino->sb->secLength;
        end=((file->pos%blocksize+count)>blocksize ? blocksize-1 : (file->pos%blocksize+count))/file->ino->sb->secLength;
        readBlock(file->ino->sb,block,buffer,start,end);
      }
      nextblockpos=(file->pos/blocksize)*blocksize+blocksize;
      findblock=0;
    }
    if (file->pos<nextblockpos)
    {
      if (extent==-1) *buf++='\0'; else *buf++=buffer[file->pos%blocksize];
      ++file->pos;
      ++got;
      --count;
    }
    else if (file->pos==nextextpos) findext=1; else findblock=1;
  }
#ifdef CPMFS_DEBUG
  fprintf(stderr,"cpmRead: read %d bytes, now at position %ld\n",got,(long)file->pos);
#endif
  return got;
}
/*}}}*/
/* cpmWrite           -- write                                   */ /*{{{*/
int cpmWrite(struct cpmFile *file, const char *buf, int count)
{
  int findext=1,findblock=1,extent=-1,extentno=-1,got=0,nextblockpos=-1,nextextpos=-1;
  int blocksize=file->ino->sb->blksiz;
  int extcap=(file->ino->sb->size<256 ? 16 : 8)*blocksize;
  int block=-1,start=-1,end=-1,ptr=-1;
  char buffer[16384];

  while (count>0)
  {
    if (findext)
    {
      extentno=file->pos/16384;
      extent=findFileExtent(file->ino->sb,file->ino->sb->dir[file->ino->ino].status,file->ino->sb->dir[file->ino->ino].name,file->ino->sb->dir[file->ino->ino].ext,0,extentno);
      nextextpos=(file->pos/extcap)*extcap+extcap;
      findext=0;
      findblock=1;
      updateTimeStamps(file->ino,extent);
    }
    if (findblock)
    {
      if (start!=-1)
      {
        int last;
    
        last=writeBlock(file->ino->sb,block,buffer,start,end);
        if (file->ino->sb->size<256) for (last=15; last>=ptr; --last)
        {
          if (file->ino->sb->dir[extent].pointers[last]) break;
        }
        else for (last=14; last>=ptr; last-=2)
        {
          if (file->ino->sb->dir[extent].pointers[last] || file->ino->sb->dir[extent].pointers[last+1]) break;
        }
        if (last==ptr) /* we wrote the last used block of this extent */
        {
          file->ino->sb->dir[extent].extnol=EXTENTL((file->pos-1)/16384);
          file->ino->sb->dir[extent].extnoh=EXTENTH((file->pos-1)/16384);
          file->ino->sb->dir[extent].blkcnt=((file->pos-1)%16384)/128+1;
          file->ino->sb->dir[extent].lrc=file->pos%128;
        }
      }
      if (extent==-1)
      {
        if ((extent=findFreeExtent(file->ino->sb))==-1) return (got==0 ? -1 : got);
        file->ino->sb->dir[extent]=file->ino->sb->dir[file->ino->ino];
        memset(file->ino->sb->dir[extent].pointers,0,16);
        file->ino->sb->dir[extent].extnol=EXTENTL(extentno);
        file->ino->sb->dir[extent].extnoh=EXTENTH(extentno);
        file->ino->sb->dir[extent].blkcnt=0;
        file->ino->sb->dir[extent].lrc=0;
      }
      ptr=(file->pos%extcap)/blocksize;
      if (file->ino->sb->size>=256) ptr*=2;
      block=file->ino->sb->dir[extent].pointers[ptr]&0xff;
      if (file->ino->sb->size>=256) block+=(file->ino->sb->dir[extent].pointers[ptr+1]&0xff)<<8;
      if (block==0)
      {
        if ((block=allocBlock(file->ino->sb))==-1) return (got==0 ? -1 : got);
        file->ino->sb->dir[extent].pointers[ptr]=block&0xff;
        if (file->ino->sb->size>=256) file->ino->sb->dir[extent].pointers[ptr+1]=(block>>8)&0xff;
        start=0;
        end=(blocksize-1)/file->ino->sb->secLength;
        memset(buffer,0,blocksize);
      }
      else
      {
        start=(file->pos%blocksize)/file->ino->sb->secLength;
        end=((file->pos%blocksize+count-1)>=blocksize ? blocksize-1 : (file->pos%blocksize+count-1))/file->ino->sb->secLength;
        if (file->pos%file->ino->sb->secLength) readBlock(file->ino->sb,block,buffer,start,start);
        if (end!=start && (file->pos+count-1)<blocksize) readBlock(file->ino->sb,block,buffer+end*file->ino->sb->secLength,end,end);
      }
      nextblockpos=(file->pos/blocksize)*blocksize+blocksize;
      findblock=0;
    }
    buffer[file->pos%blocksize]=*buf++;
    ++file->pos;
    if (file->ino->size<file->pos) file->ino->size=file->pos;
    ++got;
    if (file->pos==nextblockpos) { if (file->pos==nextextpos) findext=1; else findblock=1; }
    --count;
  }
  if (start!=-1)
  {
    int last;
    
    last=writeBlock(file->ino->sb,block,buffer,start,end);
    if (file->ino->sb->size<256) for (last=15; last>=ptr; --last)
    {
      if (file->ino->sb->dir[extent].pointers[last]) break;
    }
    else for (last=14; last>=ptr; last-=2)
    {
      if (file->ino->sb->dir[extent].pointers[last] || file->ino->sb->dir[extent].pointers[last+1]) break;
    }
    if (last==ptr) /* we wrote the last used block of this extent */
    {
      file->ino->sb->dir[extent].extnol=EXTENTL((file->pos-1)/16384);
      file->ino->sb->dir[extent].extnoh=EXTENTH((file->pos-1)/16384);
      file->ino->sb->dir[extent].blkcnt=((file->pos-1)%16384)/128+1;
      file->ino->sb->dir[extent].lrc=file->pos%128;
      writePhysDirectory(file->ino->sb);
    }
  }
  return got;
}
/*}}}*/
/* cpmClose           -- close                                   */ /*{{{*/
int cpmClose(struct cpmFile *file)
{
  if (file->mode&O_WRONLY) return (writePhysDirectory(file->ino->sb));
  return 0;
}
/*}}}*/
/* cpmCreat           -- creat                                   */ /*{{{*/
int cpmCreat(struct cpmInode *dir, const char *fname, struct cpmInode *ino, umode_t umode)
{
  int user;
  char name[8],extension[3];
  int extent;
  struct cpmSuperBlock *drive;
  struct PhysDirectoryEntry *ent;

  if (!S_ISDIR(dir->mode))
  {
    boo="No such file or directory";
    return -1;
  }
  if (splitFilename(fname,dir->sb->type,name,extension,&user)==-1) return -1;
#ifdef CPMFS_DEBUG
  fprintf(stderr,"cpmCreat: %s -> %d:%-.8s.%-.3s\n",fname,user,name,extension);
#endif
  if (findFileExtent(dir->sb,user,name,extension,0,-1)!=-1) return -1;
  drive=dir->sb;
  if ((extent=findFreeExtent(dir->sb))==-1) return -1;
  ent=dir->sb->dir+extent;
  memset(ent,0,32);
  ent->status=user;
  memcpy(ent->name,name,8);
  memcpy(ent->ext,extension,3);
  ino->ino=extent;
  ino->mode=S_IFREG|umode;
  ino->size=0;
  time(&ino->atime);
  time(&ino->mtime);
  time(&ino->ctime);
  ino->sb=dir->sb;
  updateTimeStamps(ino,extent);
  writePhysDirectory(dir->sb);
  return 0;
}
/*}}}*/
