/* #includes */ /*{{{C}}}*//*{{{*/
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
extern char *optarg;
extern int optind,opterr,optopt;
int getopt(int argc, char * const *argv, const char *optstring);

#include "cpmdir.h"
#include "cpmfs.h"
/*}}}*/

/* bcdCheck -- check format and range of BCD digit */ /*{{{*/
static int bcdCheck(int n, int max, const char *msg, const char *unit, int extent1, int extent2)
{
  if (((n>>4)&0xf)>10 || (n&0xf)>10 || (((n>>4)&0xf)*10+(n&0xf))>=max)
  {
    printf("Error: Bad %s %s (extent=%d/%d, %s=%02x)\n",msg,unit,extent1,extent2,unit,n&0xff);
    return -1;
  }
  else return 0;
}
/*}}}*/
/* fsck -- file system check */ /*{{{*/
static int fsck(struct cpmInode *root, const char *image, int type)
{
  /* variables */ /*{{{*/
  int ret=0;
  int extent,extent2;
  struct PhysDirectoryEntry *dir,*dir2;
  struct cpmSuperBlock *sb=root->sb;
  /*}}}*/

  /* Phase 1: check extent fields */ /*{{{*/
  printf("Phase 1: check extent fields\n");
  for (extent=0; extent<sb->maxdir; ++extent)
  {
    char *status;
    int usedBlocks=0;

    dir=sb->dir+extent;
    status=&dir->status;
    /* check status */ /*{{{*/
    {
      if (!(*status>=0 && *status<=(type==CPMFS_P2DOS ? 31 : 15)) && *status!=(char)0xe5 && !(type==CPMFS_P2DOS && *status==0x21) && *status!=0x20)
      {
        printf("Error: Bad status (extent=%d, status=0x%02x)\n",extent,*status&0xff);
        ret=1;
        continue;
      }
      else if (*status==(char)0xe5) continue;
    }
    /*}}}*/
    if (*status>=0 && *status<=(type==CPMFS_P2DOS ? 31 : 15))
    {
      /* check name and extension */ /*{{{*/
      {
        int i,j;
        char *c;

        for (i=0; i<8; ++i)
        {
          c=&(dir->name[i]);
          if (!ISFILECHAR(i,*c&0x7f) || islower(*c&0x7f))
          {
            printf("Error: Bad name (extent=%d, name=\"",extent);
            ret=1;
            for (j=0; j<8; ++j)
            {
              c=&(dir->name[j]);
              if ((*c&0x7f)<' ')
              {
                putchar('\\'); putchar('0');
                putchar('0'+((*c>>3)&0x03));
                putchar('0'+(*c&0x03));
              }
              else putchar(*c&0x7f);
            }
            printf("\", position=%d)\n",i);
            break;
          }
        }
        for (i=0; i<3; ++i)
        {
          c=&(dir->ext[i]);
          if (!ISFILECHAR(1,*c&0x7f) || islower(*c&0x7f))
          {
            printf("Error: Bad extension (extent=%d, extension=\"",extent);
            ret=1;
            for (j=0; j<3; ++j)
            {
              c=&(dir->ext[j]);
              if ((*c&0x7f)<' ')
              {
                putchar('\\'); putchar('0');
                putchar('0'+((*c>>3)&0x03));
                putchar('0'+(*c&0x03));
              }
              else putchar(*c&0x7f);
            }
            printf("\", position=%d)\n",i);
            break;
          }
        }
      }
      /*}}}*/
      /* check extent number */ /*{{{*/
      if ((dir->extnol&0xff)>0x1f)
      {
        printf("Error: Bad lower bits of extent number (extent=%d, low bits=%d)\n",extent,dir->extnol&0xff);
        ret=1;
      }
      if ((dir->extnoh&0xff)>0x3f)
      {
        printf("Error: Bad higher bits of extent number (extent=%d, high bits=%d)\n",extent,dir->extnoh&0xff);
        ret=1;
      }
      /*}}}*/
      /* check last record byte count */ /*{{{*/
      if ((dir->lrc&0xff)>128)
      {
        printf("Error: Bad last record byte count (extent=%d, lrc=%d)\n",extent,dir->lrc&0xff);
        ret=1;
      }
      /*}}}*/
      /* check block number range */ /*{{{*/
      {
        int block,min,max,i;

        min=(sb->maxdir*32+sb->blksiz-1)/sb->blksiz;
        max=sb->size;
        for (i=0; i<16; ++i)
        {
          block=dir->pointers[i]&0xff;
          if (sb->size>=256) block+=(dir->pointers[++i]&0xff)<<8;
          if (block>0)
          {
            ++usedBlocks;
            if (block<min || block>max)
            {
              printf("Error: Bad block number (extent=%d, block=%d)\n",extent,block);
              ret=1;
            }
          }
        }
      }
      /*}}}*/
      /* check number of used blocks ? */ /*{{{*/
      /*}}}*/
      /* check record count */ /*{{{*/
      {
        int min,max,records;

        if (usedBlocks==0)
        {
          min=0;
          max=0;
        }
        else
        {
          min=(sb->blksiz/128)*(usedBlocks-1)+1;
          max=(sb->blksiz/128)*usedBlocks;
        }
        records=dir->blkcnt&0xff;
        if (records<min || records>max)
        {
          printf("Error: Bad record count (extent=%d, record count=%d)\n",extent,records);
          ret=1;
        }
      }
      /*}}}*/
      /* check for too large .com files */ /*{{{*/
      if (((EXTENT(dir->extnol,dir->extnoh)==3 && dir->blkcnt>=126) || EXTENT(dir->extnol,dir->extnoh)>=4) && (dir->ext[0]&0x7f)=='C' && (dir->ext[1]&0x7f)=='O' && (dir->ext[2]&0x7f)=='M')
      {
        int i;
      
        printf("Warning: Oversized .COM file (extent=%d, name=\"",extent);
        for (i=0; i<8; ++i)
        {
          char c;

          c=dir->name[i];
          if ((c&0x7f)<' ')
          {
            putchar('\\'); putchar('0');
            putchar('0'+((c>>3)&0x03));
            putchar('0'+(c&0x03));
          }
          else putchar(c&0x7f);
        }
        printf("\")\n");
      }
      /*}}}*/
    }
    else if (*status==33) /* check time stamps ? */ /*{{{*/
    {
      unsigned long modified,created;

      if (sb->dir[extent2=(extent&~3)].status!=(char)0xe5) /* time stamps for first of the three extents */ /*{{{*/
      {
        bcdCheck(dir->name[2],24,"modification date","hour",extent,extent2);
        bcdCheck(dir->name[3],60,"modification date","minute",extent,extent2);
        bcdCheck(dir->name[6],24,"creation date","hour",extent,extent2);
        bcdCheck(dir->name[7],60,"creation date","minute",extent,extent2);
        modified=(dir->name[0]+(dir->name[1]<<8))*(0x60*0x60)+dir->name[2]*0x60+dir->name[3];
        created=(dir->name[4]+(dir->name[5]<<8))*(0x60*0x60)+dir->name[6]*0x60+dir->name[7];
        if (modified<created)
        {
          printf("Error: Modification date earlier than creation date (extent=%d/%d)\n",extent,extent2);
          ret=1;
        }
      }
      /*}}}*/
      if (sb->dir[extent2=(extent&~3)+1].status!=(char)0xe5) /* time stamps for second */ /*{{{*/
      {
        bcdCheck(dir->lrc,24,"modification date","hour",extent,extent2);
        bcdCheck(dir->extnoh,60,"modification date","minute",extent,extent2);
        bcdCheck(dir->pointers[1],24,"creation date","hour",extent,extent2);
        bcdCheck(dir->pointers[2],60,"creation date","minute",extent,extent2);
        modified=(dir->ext[2]+(dir->extnol<<8))*(0x60*0x60)+dir->lrc*0x60+dir->extnoh;
        created=(dir->blkcnt+(dir->pointers[0]<<8))*(0x60*0x60)+dir->pointers[1]*0x60+dir->pointers[2];
        if (modified<created)
        {
          printf("Error: Modification date earlier than creation date (extent=%d/%d)\n",extent,extent2);
          ret=1;
        }
      }
      /*}}}*/
      if (sb->dir[extent2=(extent&~3)+2].status!=(char)0xe5) /* time stamps for third */ /*{{{*/
      {
        bcdCheck(dir->pointers[7],24,"modification date","hour",extent,extent2);
        bcdCheck(dir->pointers[8],60,"modification date","minute",extent,extent2);
        bcdCheck(dir->pointers[11],24,"creation date","hour",extent,extent2);
        bcdCheck(dir->pointers[12],60,"creation date","minute",extent,extent2);
        modified=(dir->pointers[5]+(dir->pointers[6]<<8))*(0x60*0x60)+dir->pointers[7]*0x60+dir->pointers[8];
        created=(dir->pointers[9]+(dir->pointers[10]<<8))*(0x60*0x60)+dir->pointers[11]*0x60+dir->pointers[12];
        if (modified<created)
        {
          printf("Error: Modification date earlier than creation date (extent=%d/%d)\n",extent,extent2);
          ret=1;
        }
      }
      /*}}}*/
    }
    /*}}}*/
  }
  /*}}}*/
  /* Phase 2: check extent connectivity */ /*{{{*/
  printf("Phase 2: check extent connectivity\n");
  /* check multiple allocated blocks */ /*{{{*/
  for (extent=0; extent<sb->maxdir; ++extent) if ((dir=sb->dir+extent)->status>=0 && dir->status<=(type==CPMFS_P2DOS ? 31 : 15))
  {
    int i,j,block,block2;

    for (i=0; i<16; ++i)
    {
      block=dir->pointers[i]&0xff;
      if (sb->size>=256) block+=(dir->pointers[++i]&0xff)<<8;
      for (extent2=0; extent2<sb->maxdir; ++extent2) if ((dir2=sb->dir+extent2)->status>=0 && dir2->status<=(type==CPMFS_P2DOS ? 31 : 15))
      {
        for (j=0; j<16; ++j)
        {
          block2=dir2->pointers[j]&0xff;
          if (sb->size>=256) block2+=(dir2->pointers[++j]&0xff)<<8;
          if (block!=0 && block2!=0 && block==block2 && !(extent==extent2 && i==j))
          {
            printf("Error: Multiple allocated block (extent=%d,%d, block=%d)\n",extent,extent2,block);
            ret=1;
          }
        }
      }
    }
  }
  /*}}}*/
  /* check multiple extents */ /*{{{*/
  for (extent=0; extent<sb->maxdir; ++extent) if ((dir=sb->dir+extent)->status>=0 && dir->status<=(type==CPMFS_P2DOS ? 31 : 15))
  {
    for (extent2=0; extent2<sb->maxdir; ++extent2) if ((dir2=sb->dir+extent2)->status>=0 && dir2->status<=(type==CPMFS_P2DOS ? 31 : 15))
    {
      if (extent!=extent2 && EXTENT(dir->extnol,dir->extnoh)==EXTENT(dir2->extnol,dir2->extnoh) && dir->status==dir2->status)
      {
        int i;

        for (i=0; i<8 && (dir->name[i]&0x7f)==(dir2->name[i]&0x7f); ++i);
        if (i==8)
        {
          for (i=0; i<3 && (dir->ext[i]&0x7f)==(dir2->ext[i]&0x7f); ++i);
          if (i==3)
          {
            printf("Error: Duplicate extent (extent=%d,%d)\n",extent,extent2);
            ret=1;
          }
        }
      }
    }
  }
  /*}}}*/
  /*}}}*/
  if (ret==0) /* print statistics */ /*{{{*/
  {
    struct cpmStatFS statfsbuf;
    int fragmented=0,borders=0;

    cpmStatFS(root,&statfsbuf);
    for (extent=0; extent<sb->maxdir; ++extent) if ((dir=sb->dir+extent)->status>=0 && dir->status<=(type==CPMFS_P2DOS ? 31 : 15))
    {
      int i,block,previous=-1;

      for (i=0; i<16; ++i)
      {
        block=dir->pointers[i]&0xff;
        if (sb->size>=256) block+=(dir->pointers[++i]&0xff)<<8;
        if (previous!=-1)
        {
          if (block!=0 && block!=(previous+1)) ++fragmented;
          ++borders;
        }
        previous=block;
      }
    }
    fragmented=(borders ? (1000*fragmented)/borders : 0);
    printf("%s: %ld/%ld files (%d.%d%% non-contigous), %ld/%ld blocks\n",image,statfsbuf.f_files-statfsbuf.f_ffree,statfsbuf.f_files,fragmented/10,fragmented%10,statfsbuf.f_blocks-statfsbuf.f_bfree,statfsbuf.f_blocks);
  }
  /*}}}*/
  return ret;
}
/*}}}*/

const char cmd[]="fsck.cpm";

/* main */ /*{{{*/
int main(int argc, char *argv[])
{
  char *image;
  const char *format=FORMAT;
  int c,usage=0,type=CPMFS_DR22;
  struct cpmSuperBlock sb;
  struct cpmInode root;

  while ((c=getopt(argc,argv,"f:ph?"))!=EOF) switch(c)
  {
    case 'f': format=optarg; break;
    case 'p': type=CPMFS_P2DOS; break;
    case 'h':
    case '?': usage=1; break;
  }

  if (optind!=(argc-1)) usage=1;
  else image=argv[optind];

  if (usage)
  {
    fprintf(stderr,"Usage: %s [-f format] [-p] image\n",cmd);
    exit(1);
  }
  if ((sb.fd=open(image,O_RDONLY))==-1) 
  {
    fprintf(stderr,"%s: can not open %s: %s\n",cmd,image,strerror(errno));
    exit(1);
  }
  getformat(format,type,&sb,&root);
  exit(fsck(&root,image,type));
}
/*}}}*/
